import logging

from kombu import Connection, Exchange

from twingly_pyamqp.amqp_config import AMQPconfig
from twingly_pyamqp.constants import DEFAULT_EXCHANGE_OPTS


class Publisher:
    def __init__(
        self,
        exchange_name: str | None = None,
        routing_key: str | None = None,
        config: AMQPconfig = None,
        exchange_opts: dict | None = None,
        max_retries: int | None = None,
        heartbeat: int = 60,
    ):
        self.logger = logging.getLogger(__name__)
        self.config = config or AMQPconfig()
        self.connection = Connection(self.config.connection_urls(), heartbeat=heartbeat)
        self.exchange = Exchange(
            exchange_name or "", **(exchange_opts or DEFAULT_EXCHANGE_OPTS)
        )
        self.routing_key = routing_key
        self.max_retries = max_retries

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def publish(
        self,
        payload: object,
        routing_key: str | None = None,
        publish_args: dict | None = None,
    ):
        key = routing_key or self.routing_key

        producer = self.connection.Producer()
        publish = self.connection.ensure(
            producer,
            producer.publish,
            errback=self.on_connection_error,
            max_retries=self.max_retries,
        )
        publish(
            payload,
            exchange=self.exchange,
            routing_key=key,
            **(publish_args or {}),
        )

    def on_connection_error(self, exc: Exception, _):
        self.logger.error(f"Connection error occurred: {exc}")

    def close(self):
        self.connection.close()
        self.logger.info("Publisher closed.")



import os

from dotenv import load_dotenv

load_dotenv()


class AMQPconfig:
    def __init__(
        self,
        rabbitmq_hosts: str | None = None,
        rabbitmq_port: int | None = None,
        amqp_user: str | None = None,
        amqp_password: str | None = None,
        *,
        ssl: bool = False,
    ):
        self.ssl = ssl
        hosts_env = rabbitmq_hosts or os.getenv("RABBITMQ_N_HOST", "localhost")
        self.hosts = [h.strip() for h in hosts_env.split(",")]

        self.rabbitmq_port = rabbitmq_port or int(os.getenv("RABBITMQ_PORT", "5672"))
        self.amqp_user = amqp_user or os.getenv("AMQP_USERNAME", "guest")
        self.amqp_password = amqp_password or os.getenv("AMQP_PASSWORD", "guest")

    def connection_urls(self) -> str:
        scheme = "amqps" if self.ssl else "amqp"
        urls = [
            f"{scheme}://{self.amqp_user}:{self.amqp_password}@{host}:{self.rabbitmq_port}"
            for host in self.hosts
        ]
        return ";".join(urls)



import logging
import threading
from collections.abc import Callable

from kombu import Connection, Exchange, Queue, binding
from kombu.mixins import ConsumerMixin

from twingly_pyamqp.amqp_config import AMQPconfig
from twingly_pyamqp.constants import DEFAULT_EXCHANGE_OPTS, DEFAULT_QUEUE_OPTS


class Subscription(ConsumerMixin):
    def __init__(
        self,
        queue_names: list[str],
        exchange_name: str | None = None,
        bindings: dict[str, list[str]] | None = None,
        config: AMQPconfig = None,
        queue_opts: dict | None = None,
        exchange_opts: dict | None = None,
        heartbeat: int = 30,
    ):
        if not isinstance(queue_names, list):
            msg = "queue_names must be a list of strings"
            raise TypeError(msg)

        if not queue_names:
            msg = "Must provide at least one queue name"
            raise ValueError(msg)

        if not all(isinstance(queue, str) for queue in queue_names):
            msg = "All queue names must be strings"
            raise TypeError(msg)

        self.logger = logging.getLogger(__name__)
        self.config = config or AMQPconfig()
        self.thread = None
        self.connection = Connection(
            self.config.connection_urls(), heartbeat=heartbeat, ssl=self.config.ssl
        )
        if queue_opts is None:
            queue_opts = DEFAULT_QUEUE_OPTS
        if exchange_opts is None:
            exchange_opts = DEFAULT_EXCHANGE_OPTS

        self.exchange = Exchange(exchange_name or "", **exchange_opts)
        self.bindings = self._get_bindings(bindings)

        self.queues = [
            Queue(
                name,
                exchange=self.exchange,
                bindings=self.bindings.get(name),
                **queue_opts,
            )
            for name in queue_names
        ]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get_consumers(self, Consumer, _):  # noqa: N803 - Consumer arg is a class required by ConsumerMixin
        return [
            Consumer(queues=self.queues, callbacks=self.callbacks, **self.consumer_args)
        ]

    def consume(self, *args, **kwargs):
        consume = self.connection.ensure(self.connection, super().consume)
        return consume(*args, **kwargs)

    def subscribe(
        self,
        callbacks: list[Callable[[str, object], None]],
        consumer_args: dict | None = None,
        *,
        blocking: bool = True,
    ) -> None:
        if not callbacks:
            msg = "Must provide at least one callback"
            raise ValueError(msg)
        if self.thread and self.thread.is_alive():
            msg = "Consumer is already running; call cancel() before subscribing again"
            raise RuntimeError(msg)

        self.callbacks = callbacks
        self.consumer_args = consumer_args or {}

        if blocking:
            self.run()
        else:
            self.thread = threading.Thread(target=self.run, daemon=True)
            self.thread.start()

    def cancel(self) -> None:
        self.should_stop = True

        if self.thread:
            self.thread.join()
            self.thread = None

        self.should_stop = False

    def purge_queue(self, queue_name: str):
        self.logger.info(f"Purging queue: {queue_name}")
        self.connection.channel().queue_purge(queue=queue_name)

    def _get_bindings(self, bindings):
        if not bindings:
            return {}

        return {
            queue_name: [binding(self.exchange, routing_key=rk) for rk in routing_keys]
            for queue_name, routing_keys in bindings.items()
        }

    def close(self):
        self.cancel()
        self.connection.close()
        self.logger.info("Subscription closed.")
